
#ifndef WEIF_EXPORT_H
#define WEIF_EXPORT_H

#ifdef WEIF_STATIC_DEFINE
#  define WEIF_EXPORT
#  define WEIF_NO_EXPORT
#else
#  ifndef WEIF_EXPORT
#    ifdef weif_EXPORTS
        /* We are building this library */
#      define WEIF_EXPORT 
#    else
        /* We are using this library */
#      define WEIF_EXPORT 
#    endif
#  endif

#  ifndef WEIF_NO_EXPORT
#    define WEIF_NO_EXPORT 
#  endif
#endif

#ifndef WEIF_DEPRECATED
#  define WEIF_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef WEIF_DEPRECATED_EXPORT
#  define WEIF_DEPRECATED_EXPORT WEIF_EXPORT WEIF_DEPRECATED
#endif

#ifndef WEIF_DEPRECATED_NO_EXPORT
#  define WEIF_DEPRECATED_NO_EXPORT WEIF_NO_EXPORT WEIF_DEPRECATED
#endif

/* NOLINTNEXTLINE(readability-avoid-unconditional-preprocessor-if) */
#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef WEIF_NO_DEPRECATED
#    define WEIF_NO_DEPRECATED
#  endif
#endif

#endif /* WEIF_EXPORT_H */
