#ifndef XTENSOR
#define XTENSOR

#include <xtensor/chunk/xchunked_array.hpp>
#include <xtensor/chunk/xchunked_assign.hpp>
#include <xtensor/chunk/xchunked_view.hpp>
#include <xtensor/containers/xadapt.hpp>
#include <xtensor/containers/xarray.hpp>
#include <xtensor/containers/xbuffer_adaptor.hpp>
#include <xtensor/containers/xcontainer.hpp>
#include <xtensor/containers/xfixed.hpp>
#include <xtensor/containers/xscalar.hpp>
#include <xtensor/containers/xstorage.hpp>
#include <xtensor/containers/xtensor.hpp>
#include <xtensor/core/xaccessible.hpp>
#include <xtensor/core/xassign.hpp>
#include <xtensor/core/xeval.hpp>
#include <xtensor/core/xexpression.hpp>
#include <xtensor/core/xexpression_traits.hpp>
#include <xtensor/core/xfunction.hpp>
#include <xtensor/core/xiterable.hpp>
#include <xtensor/core/xiterator.hpp>
#include <xtensor/core/xlayout.hpp>
#include <xtensor/core/xmath.hpp>
#include <xtensor/core/xmultiindex_iterator.hpp>
#include <xtensor/core/xnoalias.hpp>
#include <xtensor/core/xoperation.hpp>
#include <xtensor/core/xsemantic.hpp>
#include <xtensor/core/xshape.hpp>
#include <xtensor/core/xstrides.hpp>
#include <xtensor/core/xtensor_config.hpp>
#include <xtensor/core/xtensor_forward.hpp>
#include <xtensor/core/xvectorize.hpp>
#include <xtensor/generators/xbuilder.hpp>
#include <xtensor/generators/xgenerator.hpp>
#include <xtensor/generators/xrandom.hpp>
#include <xtensor/io/xcsv.hpp>
#include <xtensor/io/xinfo.hpp>
#include <xtensor/io/xio.hpp>
#include <xtensor/misc/xcomplex.hpp>
#include <xtensor/misc/xfft.hpp>
#include <xtensor/misc/xhistogram.hpp>
#include <xtensor/misc/xmanipulation.hpp>
#include <xtensor/misc/xpad.hpp>
#include <xtensor/misc/xset_operation.hpp>
#include <xtensor/misc/xsort.hpp>
#include <xtensor/optional/xoptional.hpp>
#include <xtensor/optional/xoptional_assembly.hpp>
#include <xtensor/optional/xoptional_assembly_base.hpp>
#include <xtensor/optional/xoptional_assembly_storage.hpp>
#include <xtensor/reducers/xaccumulator.hpp>
#include <xtensor/reducers/xblockwise_reducer.hpp>
#include <xtensor/reducers/xblockwise_reducer_functors.hpp>
#include <xtensor/reducers/xnorm.hpp>
#include <xtensor/reducers/xreducer.hpp>
#include <xtensor/utils/xexception.hpp>
#include <xtensor/utils/xtensor_simd.hpp>
#include <xtensor/utils/xutils.hpp>
#include <xtensor/views/xaxis_iterator.hpp>
#include <xtensor/views/xaxis_slice_iterator.hpp>
#include <xtensor/views/xbroadcast.hpp>
#include <xtensor/views/xdynamic_view.hpp>
#include <xtensor/views/xfunctor_view.hpp>
#include <xtensor/views/xindex_view.hpp>
#include <xtensor/views/xmasked_view.hpp>
#include <xtensor/views/xoffset_view.hpp>
#include <xtensor/views/xrepeat.hpp>
#include <xtensor/views/xslice.hpp>
#include <xtensor/views/xstrided_view.hpp>
#include <xtensor/views/xstrided_view_base.hpp>
#include <xtensor/views/xview.hpp>
#include <xtensor/views/xview_utils.hpp>

#endif
